@@ -2,42 +2,45 @@ from __future__ import annotations
 
 import copy
 import operator
+from collections.abc import Iterable
 from functools import reduce
+from typing import Any
 
-from tatsu.util import asjson
-
-from .util import is_list
+from .infos import ParseInfo
+from .util import asjson, is_list
 
 
-class AST(dict):
+class AST(dict[str, Any]):
     _frozen = False
 
-    def __init__(self, *args, **kwargs):
+    def __init__(self, *args: Any, **kwargs: Any):
         super().__init__()
         self.update(*args, **kwargs)
         self._frozen = True
 
     @property
-    def frozen(self):
+    def frozen(self) -> bool:
         return self._frozen
 
     @property
-    def parseinfo(self):
+    def parseinfo(self) -> ParseInfo | None:
         try:
             return super().__getitem__('parseinfo')
         except KeyError:
-            pass
+            return None
 
-    def set_parseinfo(self, value):
+    def set_parseinfo(self, value: ParseInfo | None) -> None:
+        if value is None:
+            return
         super().__setitem__('parseinfo', value)
 
-    def copy(self):
+    def copy(self) -> AST:
         return copy.copy(self)
 
-    def asjson(self):
+    def asjson(self) -> Any:
         return asjson(self)
 
-    def _set(self, key, value, force_list=False):
+    def _set(self, key: str, value: Any, force_list: bool = False) -> None:
         key = self._safekey(key)
         previous = self.get(key)
 
@@ -52,13 +55,13 @@ class AST(dict):
 
         super().__setitem__(key, value)
 
-    def _setlist(self, key, value):
-        return self._set(key, value, force_list=True)
+    def _setlist(self, key: str, value: list[Any]) -> None:
+        self._set(key, value, force_list=True)
 
-    def __copy__(self):
+    def __copy__(self) -> AST:
         return AST(self)
 
-    def __getitem__(self, key):
+    def __getitem__(self, key: str) -> Any:
         if key in self:
             return super().__getitem__(key)
         key = self._safekey(key)
@@ -66,14 +69,14 @@ class AST(dict):
             return super().__getitem__(key)
         return None
 
-    def __setitem__(self, key, value):
+    def __setitem__(self, key: str, value: Any) -> None:
         self._set(key, value)
 
-    def __delitem__(self, key):
+    def __delitem__(self, key: str) -> None:
         key = self._safekey(key)
         super().__delitem__(key)
 
-    def __setattr__(self, name, value):
+    def __setattr__(self, name: str, value: Any) -> None:
         if self._frozen and name not in vars(self):
             raise AttributeError(
                 f'{type(self).__name__} attributes are fixed. '
@@ -81,7 +84,7 @@ class AST(dict):
             )
         super().__setattr__(name, value)
 
-    def __getattr__(self, name):
+    def __getattr__(self, name: str) -> Any:
         key = self._safekey(name)
         if key in self:
             return self[key]
@@ -93,7 +96,7 @@ class AST(dict):
         except AttributeError:
             return None
 
-    def __hasattribute__(self, name):
+    def __hasattribute__(self, name: str) -> bool:
         try:
             super().__getattribute__(name)
         except (TypeError, AttributeError):
@@ -101,15 +104,15 @@ class AST(dict):
         else:
             return True
 
-    def __reduce__(self):
+    def __reduce__(self) -> tuple[Any, Any]:
         return (AST, (list(self.items()),))
 
-    def _safekey(self, key):
+    def _safekey(self, key: str) -> str:
         while self.__hasattribute__(key):
             key += '_'
         return key
 
-    def _define(self, keys, list_keys=None):
+    def _define(self, keys: Iterable[str], list_keys: Iterable[str] | None = None):
         for key in (self._safekey(k) for k in keys):
             if key not in self:
                 super().__setitem__(key, None)
@@ -118,19 +121,20 @@ class AST(dict):
             if key not in self:
                 super().__setitem__(key, [])
 
-    def __json__(self, seen=None):
+    def __json__(self, seen: set[int] | None = None) -> Any:
         return {name: asjson(value, seen=seen) for name, value in self.items()}
 
-    def __repr__(self):
+    def __repr__(self) -> str:
         return repr(self.asjson())
 
-    def __str__(self):
+    def __str__(self) -> str:
         return str(self.asjson())
 
-    def __hash__(self):
-        # NOTE: objects are actually mutable during creation
-        return reduce(
-            operator.xor,
-            (hash((name, id(value))) for name, value in self.items()),
-            0,
+    def __hash__(self) -> int:  # type: ignore
+        return hash(
+            reduce(
+                operator.xor,
+                (hash((name, id(value))) for name, value in self.items()),
+                0,
+            ),
         )
