@@ -4,17 +4,17 @@ from typing import Any
 from .bootstrap import EBNFBootstrapParser
 from .buffering import Buffer
 from .grammars import PRAGMA_RE
+from .infos import ParserConfig
 from .parser_semantics import EBNFGrammarSemantics
-from .parserconfig import ParserConfig
 from .semantics import ASTSemantics
 
 
 class EBNFBuffer(Buffer):
     def __init__(
         self,
-        text,
+        text: str,
         /,
-        filename=None,
+        filename: str | None = None,
         config: ParserConfig | None = None,
         **settings: Any,
     ):
